/*
 this software is copyright (c) 2001 Dominik Behr <dominikbehr@yahoo.com>
 this software is distributed under GNU Lesser General Public License
 see LICENSE.txt for details 
*/
#include <windows.h>
#define GXDLL_EXPORTS
#include "gx.h"
#include "debug.h"

/*
 according to Vrc4171a documentation framebuffer is located at 0x0A000000L
 but when I mapped this area and started writting there my LCD became white
 then black and after reset sound stopped working. then after hard reset
 sound wasnt working too. GULP! removing batteries helped. :)
 so I have set screen using GDI to some pattern and mapped whole memory megabyte 
 by megabyte and scanned for the pattern. and found this address :) 
*/
#define CASIO_FRAMEBUFFER 0x0A200000L

/* 
   VirtualCopy is the magic function which allows booting Linux on Windows CE 
   devices.

   http://msdn.microsoft.com/library/default.asp?url=/library/en-us/wceoem/htm/krnl_f_33.asp

   there is one KB article about it
   http://support.microsoft.com/support/kb/articles/Q213/3/59.ASP

   
   parameters are:
   LPVOID pMemoryBlock
   LPVOID pPhysicalAddress
   DWORD dwBlockSize
   DWORD dwProtectionFlags
   
   it seems there are two ways to map memory
   if physical address is less than 512MB it should be ORed with 0x80000000L
   and ProtectionFlags should be PAGE_READWRITE | PAGE_NOCACHE
   else
   physical address has to be shifted right by 8 abd
   ProtectionFlags should be PAGE_READWRITE | PAGE_NOCACHE | PAGE_PHYSICAL
*/   
extern "C" {
__declspec(dllimport) BOOL VirtualCopy(LPVOID, LPVOID, DWORD, DWORD);
};


/***********************************************************************************************

***********************************************************************************************/
BOOL APIENTRY 
DllMain(HANDLE hModule, 
        DWORD  ul_reason_for_call, 
        LPVOID lpReserved)
{
 return TRUE;
}

/***********************************************************************************************

***********************************************************************************************/
static DWORD
GetPageSize(void)
{
 static int init = 0;
 static SYSTEM_INFO stInfo;

 if (!init) 
    {
     GetSystemInfo(&stInfo);
     init = 1;
     }
 return (stInfo.dwPageSize);
}

static HWND hFindResult;

/***********************************************************************************************

***********************************************************************************************/
static BOOL CALLBACK 
FindEnumWindowsProc(HWND hWnd, LPARAM lParam)
{
 TCHAR aClass[256];
 GetClassName(hWnd, aClass, 256);
 if (_tcscmp(aClass, (TCHAR *)lParam) == 0)
    {
     hFindResult = hWnd;
     return FALSE;
    }
 return TRUE;     
}

/***********************************************************************************************

***********************************************************************************************/
static HWND
FindWindowByClass(LPCTSTR pszClass)
{
 hFindResult = 0;
 EnumWindows(FindEnumWindowsProc, (LPARAM)pszClass);
 return hFindResult;
}

/***********************************************************************************************

***********************************************************************************************/
static void
TaskbarHide(void)
{
 HWND hWnd;
 hWnd = FindWindowByClass(TEXT("HHTaskBar"));
 ShowWindow(hWnd, SW_HIDE);

}

/***********************************************************************************************

***********************************************************************************************/
static void
TaskbarShow(void)
{
 HWND hWnd;
 hWnd = FindWindowByClass(TEXT("HHTaskBar"));
 ShowWindow(hWnd, SW_SHOWNA);

}

//typedef LRESULT (CALLBACK *PFNWNDPROC)(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
typedef int (__cdecl *PFNWNDPROC)(void);

#define HOTKEY_A 10
#define HOTKEY_B 11
#define HOTKEY_S 12
#define HOTKEY_V 13

static char *g_pMem = 0;
static unsigned int g_uViewportTop = 0;
static HWND g_hWnd = 0;
static DWORD g_dwFlags;
static DWORD g_dwButtonA = VK_RETURN; /* calendar */
static DWORD g_dwButtonB = VK_SPACE; /* rolodex */
static DWORD g_dwButtonS = VK_TAB; /* rocket */
static DWORD g_dwButtonV = 'V'; /* voice recording */
static PFNWNDPROC pfnOldWndProc = 0;

/***********************************************************************************************

***********************************************************************************************/
LRESULT CALLBACK 
FilterProc(HWND hWnd, 
           UINT uMsg, 
           WPARAM wParam, 
           LPARAM lParam)
{
 if (uMsg == WM_HOTKEY)
    {
     DWORD dwButton;
     DebugMessage(TEXT("HOTKEY %08X,%08X,%08X,%08X\n"),
                  hWnd, uMsg, wParam, lParam);
     if (wParam == HOTKEY_A)
        dwButton = g_dwButtonA;
     else
     if (wParam == HOTKEY_B)
        dwButton = g_dwButtonB;
     else
     if (wParam == HOTKEY_S)
        dwButton = g_dwButtonS;
     else
     if (wParam == HOTKEY_V)
        dwButton = g_dwButtonV;
     else
        return CallWindowProc(pfnOldWndProc, hWnd, uMsg, wParam, lParam);
     CallWindowProc(pfnOldWndProc, hWnd, WM_KEYDOWN, (WPARAM)dwButton, (LPARAM)0x00000001);
     CallWindowProc(pfnOldWndProc, hWnd, WM_KEYUP, (WPARAM)dwButton, (LPARAM)0xC0000001);
     return 0;
    }
 else
 if (uMsg == WM_DESTROY)
    {
     SetWindowLong(hWnd, GWL_WNDPROC, (LONG)pfnOldWndProc);
     return CallWindowProc(pfnOldWndProc, hWnd, uMsg, wParam, lParam);
    }
 else
    return CallWindowProc(pfnOldWndProc, hWnd, uMsg, wParam, lParam);
}

/***********************************************************************************************

***********************************************************************************************/
GXDLL_API int 
GXOpenDisplay(HWND hWnd, DWORD dwFlags)
{
 DWORD dwPageSize;
 DWORD dwSize;

 g_dwFlags = dwFlags;
 if (hWnd)
    g_hWnd = hWnd;
 else
    g_hWnd = GetForegroundWindow();

 dwPageSize = GetPageSize();
 dwSize = 256*320*2;
 dwSize += dwPageSize - 1;
 dwSize &= 0xFFFFFFFF - (dwPageSize - 1);
 g_pMem = (char *)VirtualAlloc(0, dwSize, MEM_RESERVE, PAGE_NOACCESS);
 if (!g_pMem)
    return 0;

 VirtualCopy((LPVOID)g_pMem, 
             (LPVOID)((CASIO_FRAMEBUFFER | 0x80000000L)), 
             dwSize, 
             PAGE_READWRITE | PAGE_NOCACHE);

 if (g_dwFlags & GX_FULLSCREEN)
    TaskbarHide();
     
 return 1;
}

/***********************************************************************************************

***********************************************************************************************/
GXDLL_API int 
GXCloseDisplay()
{
 if (!g_pMem)
    return 0;

 if (g_dwFlags & GX_FULLSCREEN)
    TaskbarShow();

 VirtualFree((LPVOID)g_pMem, 0, MEM_RELEASE);
 g_pMem = 0;
 return 1;
}

/***********************************************************************************************

***********************************************************************************************/
GXDLL_API void *
GXBeginDraw(void)
{
 return (void *)(g_pMem + g_uViewportTop * 512);
}

/***********************************************************************************************

***********************************************************************************************/
GXDLL_API int 
GXEndDraw(void)
{
 return 1;
}

/***********************************************************************************************

***********************************************************************************************/
static DWORD 
SetButtonCode(LPCTSTR pszKey,
              DWORD dwValue)
{
 HKEY hKey;
 DWORD dwType;
 DWORD dwRes;
 DWORD dwSize;
 LONG lR;

 lR =
 RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
              pszKey,
              0,
              0,
              &hKey);
 #ifdef DEBUG               
 DebugMessage(TEXT("Open R %08X key %08X\n"), lR, hKey);
 #endif
 dwSize = sizeof(dwRes);
 lR =
 RegQueryValueEx(hKey,
                 TEXT("AppDefs"),
                 NULL,
                 &dwType,
                 (LPBYTE)&dwRes,
                 &dwSize);
 #ifdef DEBUG
 DebugMessage(TEXT("R %08X Val %08X type %08X size %08X\n"), lR, dwRes, dwType, dwSize);
 #endif
 lR =
 RegSetValueEx(hKey,
               TEXT("AppDefs"),
               0,
               dwType,
               (LPBYTE)&dwValue,
               sizeof(dwValue));
 #ifdef DEBUG
 DebugMessage(TEXT("Set R %08X\n"), lR);
 Sleep(2*1000);
 DWORD dwTest;
 dwSize = sizeof(dwRes);
 lR =
 RegQueryValueEx(hKey,
                 TEXT("AppDefs"),
                 NULL,
                 &dwType,
                 (LPBYTE)&dwTest,
                 &dwSize);
 DebugMessage(TEXT("TEST R %08X Val %08X type %08X size %08X\n"), lR, dwTest, dwType, dwSize);
 #endif

 lR =
 RegCloseKey(hKey);
 #ifdef DEBUG
 DebugMessage(TEXT("Close R %08X\n"), lR);
 #endif
 return dwRes;
}

/***********************************************************************************************

***********************************************************************************************/
static void    
SetButtons(void)
{
 /* default mappings 
    A 195
    B 196
    S 194
    V 193
 */
 g_dwButtonA = SetButtonCode(TEXT("Drivers\\CASIO\\Keybd\\Keys\\ButtonA"), g_dwButtonA);
 g_dwButtonB = SetButtonCode(TEXT("Drivers\\CASIO\\Keybd\\Keys\\ButtonB"), g_dwButtonB);
 g_dwButtonS = SetButtonCode(TEXT("Drivers\\CASIO\\Keybd\\Keys\\ButtonS"), g_dwButtonS);
 g_dwButtonV = SetButtonCode(TEXT("Drivers\\CASIO\\Keybd\\Keys\\ButtonV"), g_dwButtonV); 
 /*
  there is little problem some stupid process restores button mappings
  after i change them
 */
}

/***********************************************************************************************

***********************************************************************************************/
GXDLL_API int 
GXOpenInput(void)
{
 BOOL bReg;

 #if 1
 bReg =
 RegisterHotKey(g_hWnd, HOTKEY_A, MOD_WIN, 195);
 #ifdef DEBUG
 DebugMessage(TEXT("RegisterHotKey %08X\n"), bReg);
 #endif
 bReg =
 RegisterHotKey(g_hWnd, HOTKEY_B, MOD_WIN, 196);
 #ifdef DEBUG
 DebugMessage(TEXT("RegisterHotKey %08X\n"), bReg);
 #endif
 bReg =
 RegisterHotKey(g_hWnd, HOTKEY_S, MOD_WIN, 194);
 #ifdef DEBUG
 DebugMessage(TEXT("RegisterHotKey %08X\n"), bReg);
 #endif
 
 #if 0
 bReg =
 RegisterHotKey(g_hWnd, HOTKEY_V, MOD_WIN, 193);
 #ifdef DEBUG
 DebugMessage(TEXT("RegisterHotKey %08X\n"), bReg);
 #endif
 #endif
 #endif

 pfnOldWndProc = (PFNWNDPROC)GetWindowLong(g_hWnd, GWL_WNDPROC);
 SetWindowLong(g_hWnd, GWL_WNDPROC, (LONG)FilterProc);
 //SetButtons(); 
 return 1;
}

/***********************************************************************************************
 Action - VK_F23 0x86
 Wheel Up,  Arrow Up is VK_UP 0x26
 Wheel Down, Arrow Down is VK_DOWN 0x28
 Arrow Left is VK_LEFT 0x25
 Arrow Right is VK_RIGHT 0x27
 Exit is VK_F24 0x87


***********************************************************************************************/
GXDLL_API int 
GXCloseInput(void)
{
 //SetButtons();
 if (IsWindow(g_hWnd))
    {
     SetWindowLong(g_hWnd, GWL_WNDPROC, (LONG)pfnOldWndProc);
     #if 1
     UnregisterHotKey(g_hWnd, HOTKEY_A);
     UnregisterHotKey(g_hWnd, HOTKEY_B);
     UnregisterHotKey(g_hWnd, HOTKEY_S);
     #if 0
     UnregisterHotKey(g_hWnd, HOTKEY_V);
     #endif
     #endif
    }
 return 1;
}

/***********************************************************************************************

***********************************************************************************************/
GXDLL_API GXDisplayProperties 
GXGetDisplayProperties(void)
{
 struct GXDisplayProperties stProps;
 stProps.cxWidth = 240;
 stProps.cyHeight = 320;
 stProps.cbxPitch = 2;
 stProps.cbyPitch = 2*256;
 stProps.cBPP = 16;
 stProps.ffFormat = kfDirect565;
 return stProps; 
}

/***********************************************************************************************
 this is broken now
 i have to find out what PocketPC Casio devices return in this function
***********************************************************************************************/
GXDLL_API GXKeyList 
GXGetDefaultKeys(int iOptions)
{
 struct GXKeyList stList;
 stList.vkUp = VK_UP;				
 stList.ptUp.x = 0;				
 stList.ptUp.y = 0;
 stList.vkDown = VK_DOWN;
 stList.ptDown.x = 0;
 stList.ptDown.y = 0;
 stList.vkLeft = VK_LEFT;
 stList.ptLeft.x = 0;
 stList.ptLeft.y = 0;
 stList.vkRight = VK_RIGHT;
 stList.ptRight.x = 0;
 stList.ptRight.y = 0;
 stList.vkA = VK_RETURN;
 stList.ptA.x = 0;
 stList.ptA.y = 0;
 stList.vkB = VK_SPACE;
 stList.ptB.x = 0;
 stList.ptB.y = 0;
 stList.vkC = VK_TAB;
 stList.ptC.x = 0;
 stList.ptC.y = 0;
 stList.vkStart = 'S';
 stList.ptStart.x = 0;
 stList.ptStart.y = 0;
 return stList;
}

/***********************************************************************************************
 there is nothing here yet
***********************************************************************************************/
GXDLL_API int 
GXSuspend(void)
{
 return 1;
}

/***********************************************************************************************
 there is nothing here yet
***********************************************************************************************/
GXDLL_API int 
GXResume(void)
{
 return 1;
}

/***********************************************************************************************
 i am not sure what this function is doing but since it is suppsed to work
 when GXIsDisplayDRAMBuffer() returns TRUE i dont care
***********************************************************************************************/
GXDLL_API int 
GXSetViewport(DWORD dwTop, DWORD dwHeight, DWORD dwReserved1, DWORD dwReserved2)
{
 g_uViewportTop = dwTop;
 return 1;
}

/***********************************************************************************************
 it seems some devices do not expose hardware frame buffer and they return TRUE here
***********************************************************************************************/
GXDLL_API BOOL 
GXIsDisplayDRAMBuffer()
{
 return FALSE;
}

